import java.awt.*;
import java.awt.event.*;
//import java.awt.image.ImageObserver;
import java.awt.image.ImageConsumer;
import java.awt.image.ColorModel;
import javax.swing.*;
import java.io.File;
import java.net.MalformedURLException;
import java.util.List;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * @version 1.0.1
 * @author David Foster <davidfstr@gmail.com>
 */
public class ImageFilterDemo {
	// ### FILTER CONFIGURATION ###
	
	/**
	 * contains the filters that are displayed in the UI.
	 * 
	 * Supported filter types include:
	 * <ul>
	 * <li>ImageFilter</li>
	 * <li>BufferedImageFilter</li>
	 * <li>PixelBufferFilter</li>
	 * <li>PixelBufferFilter2</li> (since v1.0.1)
	 * <li>PixelFilter</li>
	 * </ul>
	 */
	private static List filterList;		// List<ImageFilter, BufferedImageFilter, PixelBufferFilter, PixelFilter>
	
	static {
		filterList = new ArrayList();
		
		/* --- FILTERS: Modify this list of filters to change what filters are displayed in the UI --- */
		filterList.add((ImageFilter) new IdentityFilter());
		filterList.add((PixelBufferFilter) new HorizontalFlipFilter());
		filterList.add((PixelFilter) new RGBInversionFilter());
		
		// since v1.0.1
		filterList.add((PixelBufferFilter2) new HorizontalContourExtractionFilter(5, false));
		filterList.add((PixelBufferFilter2) new VerticalContourExtractionFilter(5, false));
		filterList.add((PixelBufferFilter2) new PlusContourExtractionFilter(5));
	}
	
	// ### MAIN ###
	
	private static JFrame mainFrame;
	private static JFileWell inputFileWell;
	private static JButton chooseFileButton;
	private static JComboBox filterComboBox;
	private static ComponentWell inputImageView;
	private static ComponentWell outputImageView;
	private static Image inputImage;
	
	public static void main(String[] args) {
		inputFileWell = new JFileWell();
		
		chooseFileButton = new JButton("Choose...");
		chooseFileButton.setFont(JFileWell.FONT);
		chooseFileButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				JFileChooser fileChooser = new JFileChooser();
				int chooseResult = fileChooser.showOpenDialog(mainFrame);
				if (chooseResult == JFileChooser.APPROVE_OPTION) {
					setInputFile(fileChooser.getSelectedFile());
				}
			}
		});
		
		Container fileLine = new Container();
		fileLine.setLayout(new ListLayout(ListLayout.X_AXIS, ListLayout.CENTER));
		fileLine.add(makeBold(new JLabel("File: ")));
		fileLine.add(inputFileWell, Boolean.TRUE);
		fileLine.add(chooseFileButton);
		
		FilterComboBoxModel filterComboBoxModel = new FilterComboBoxModel();
		filterComboBox = new JComboBox(filterComboBoxModel);
		filterComboBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				// Update Output pane
				inputImageOrFilterChanged();
			}
		});
		
		inputImageView = new ComponentWell(null);
		
		JPanel inputPane = new JPanel();
		inputPane.setBorder(BorderFactory.createTitledBorder("Input"));
		inputPane.setLayout(new ListLayout(ListLayout.Y_AXIS, ListLayout.FILL));
		inputPane.add(fileLine);
		inputPane.add(inputImageView, Boolean.TRUE);
		
		Container filterLine = new Container();
		filterLine.setLayout(new ListLayout(ListLayout.X_AXIS, ListLayout.CENTER));
		filterLine.add(makeBold(new JLabel("Filter: ")));
		filterLine.add(filterComboBox, Boolean.TRUE);
		
		outputImageView = new ComponentWell(null);
		
		JPanel outputPane = new JPanel();
		outputPane.setBorder(BorderFactory.createTitledBorder("Output"));
		outputPane.setLayout(new ListLayout(ListLayout.Y_AXIS, ListLayout.FILL));
		outputPane.add(filterLine);
		outputPane.add(outputImageView, Boolean.TRUE);
		
		Container content = new Container();
		content.setLayout(new GridLayout(1, 2));
		content.add(inputPane);
		content.add(outputPane);
		
		// Generate event to update UI
		inputFileChanged();
		
		mainFrame = new JFrame();
		mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		mainFrame.setContentPane(content);
		mainFrame.pack();
		mainFrame.show();
	}
	
	private static JLabel makeBold(JLabel label) {
		label.setFont(label.getFont().deriveFont(Font.BOLD));
		return label;
	}
	
	// ### INTERNAL EVENTS ###
	
	private static Component INPUT_ERROR_FILE_NOT_SELECTED =
		new JDesktopMessagePane("No image file selected");
	
	private static Component INPUT_ERROR_BAD_IMAGE =
		new JDesktopMessagePane("Unable to load image!");
	
	private static Component INPUT_STATUS_LOADING_IMAGE =
		new JDesktopMessagePane("Loading image...");
	
	private static Component OUTPUT_STATUS_WAITING_FOR_INPUT =
		new JDesktopMessagePane("Waiting for input image");
	
	private static Component OUTPUT_STATUS_APPLYING_FILTER =
		new JDesktopMessagePane("Applying filter...");
	
	private static Component OUTPUT_ERROR_FILTER_NOT_SELECTED =
		new JDesktopMessagePane("No filter selected");
	
	// NOTE: must only be called from the event dispatching thread
	private static void inputFileChanged() {
		// Unload image from old input file (and update Output pane)
		inputImage = null;
		inputImageOrFilterChanged();
		
		File newInputFile = getInputFile();
		if (newInputFile == null) {
			inputImageView.setComponent(INPUT_ERROR_FILE_NOT_SELECTED);
		} else {
			inputImageView.setComponent(INPUT_STATUS_LOADING_IMAGE);
			
			final Image loadingInputImage;
			try {
				loadingInputImage = Toolkit.getDefaultToolkit().createImage(newInputFile.toURL());
			} catch (MalformedURLException e) {
				inputImageView.setComponent(INPUT_ERROR_BAD_IMAGE);
				e.printStackTrace(System.out);
				return;
			}
			
			// Setup a continuation to be called when the image is loaded
			chooseFileButton.setEnabled(false);
			waitForImageToFinishLoading(loadingInputImage, new ActionListener() {
				public void actionPerformed(ActionEvent evt) {
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							if ((loadingInputImage.getWidth(null) == -1) ||
								(loadingInputImage.getHeight(null) == -1))
							{
								// An error occurred while loading the image
								inputImageView.setComponent(INPUT_ERROR_BAD_IMAGE);
							} else {
								// Remember the loaded image
								inputImage = loadingInputImage;
								
								// Update ui to display image (as it is loaded)
								inputImageView.setComponent(new JLabel(new ImageIcon(inputImage)));
								mainFrame.pack();		// update window size
								
								// Update Output pane
								inputImageOrFilterChanged();
							}
							chooseFileButton.setEnabled(true);
						}
					});
				}
			});
		}
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void inputImageOrFilterChanged() {
		final ImageFilter selectedImageFilter = getSelectedFilterAsImageFilter();
		if (selectedImageFilter == null) {
			outputImageView.setComponent(OUTPUT_ERROR_FILTER_NOT_SELECTED);
		} else {
			if (inputImage == null) {
				outputImageView.setComponent(OUTPUT_STATUS_WAITING_FOR_INPUT);
			} else {
				outputImageView.setComponent(OUTPUT_STATUS_APPLYING_FILTER);
				
				new Thread("Image Filter") {
					public void run() {
						final Image filteredImage = selectedImageFilter.createFilteredImage(inputImage);
						SwingUtilities.invokeLater(new Runnable() {
							public void run() {
								outputImageView.setComponent(new JLabel(new ImageIcon(filteredImage)));
								mainFrame.pack();		// update window size
							}
						});
					}
				}.start();
			}
		}
	}
	
	// ### INTERNAL ACCESSORS ###
	
	// NOTE: must only be called from the event dispatching thread
	public static void setInputFile(File inputFile) {
		inputFileWell.setFile(inputFile);
		inputFileChanged();
	}
	
	public static File getInputFile() {
		return inputFileWell.getFile();
	}
	
	public static Object getSelectedFilter() {
		return filterComboBox.getSelectedItem();
	}
	
	public static ImageFilter getSelectedFilterAsImageFilter() {
		return filterAsImageFilter(getSelectedFilter());
	}
	
	// ### UTILITY ###
	
	private static void waitForImageToFinishLoading(Image img, final ActionListener finishListener) {
		/*
		ImageObserver imgObserver = new ImageObserver() {
			public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
				if ((infoflags & (ImageObserver.ALLBITS | ImageObserver.FRAMEBITS | ImageObserver.ERROR | ImageObserver.ABORT)) != 0) {
					// Done loading image
					finishListener.actionPerformed(null);
					return false;
				} else {
					return true;
				}
			}
		};
		
		if (Toolkit.getDefaultToolkit().prepareImage(img, -1, -1, imgObserver)) {
			imgObserver.imageUpdate(inputImage, ImageObserver.ALLBITS, -1, -1, -1, -1);
		}
		*/
		
		ImageConsumer imgConsumer = new ImageConsumer() {
			public void imageComplete(int status) {
				// Done loading image
				finishListener.actionPerformed(null);
			}
			
			public void setColorModel(ColorModel model) {}
			public void setDimensions(int width, int height) {}
			public void setHints(int hintflags) {}
			public void setPixels(int x, int y, int w, int h, ColorModel model, byte[] pixels, int off, int scansize) {}
			public void setPixels(int x, int y, int w, int h, ColorModel model, int[] pixels, int off, int scansize) {}
			public void setProperties(Hashtable props) {}
		};
		
		// (most robust on Windows than Toolkit.prepareImage)
		img.getSource().startProduction(imgConsumer);
	}
	
	/**
	 * Converts a <tt>filter</tt> of one of the supported filter types
	 * to an <tt>ImageFilter</tt>.
	 */
	private static ImageFilter filterAsImageFilter(Object filter) {
		if (filter == null)
			return null;
		
		if (filter instanceof PixelFilter)
			filter = new PixelFilterAsPixelBufferFilter((PixelFilter) filter);
		if (filter instanceof PixelBufferFilter)
			filter = new PixelBufferFilterAsBufferedImageFilter((PixelBufferFilter) filter);
		if (filter instanceof PixelBufferFilter2)
			filter = new PixelBufferFilter2AsBufferedImageFilter((PixelBufferFilter2) filter);
		if (filter instanceof BufferedImageFilter)
			filter = new BufferedImageFilterAsImageFilter((BufferedImageFilter) filter);
		
		if (filter instanceof ImageFilter) {
			return (ImageFilter) filter;
		} else {
			throw new IllegalArgumentException("unsupported filter type: "+filter.getClass().getName());
		}
	}
	
	// ### MODEL ###
	
	static class FilterComboBoxModel extends AbstractListModel implements ComboBoxModel {
		// ### ELEMENTS ###
		
		public Object getElementAt(int index) {
			return filterList.get(index);
		}
		
		public int getSize() {
			return filterList.size();
		}
		
		// ### COMBO BOX MODEL ###
		
		private Object selectedItem = null;
		
		public Object getSelectedItem() {
			return this.selectedItem;
		}
		
		public void setSelectedItem(Object selectedItem) {
			this.selectedItem = selectedItem;
		}
	}
}

class JDesktopMessagePane extends JLabel {
	public JDesktopMessagePane() {
		this.setHorizontalAlignment(SwingConstants.CENTER);
		this.setForeground(Color.WHITE);
		this.setBackground(Color.GRAY);
	}
	
	public JDesktopMessagePane(String text) {
		this();
		this.setText(text);
	}
	
	// ### IMPL ###
	
	public void paintComponent(Graphics g) {
		g.setColor(this.getBackground());
		g.fillRect(0, 0, this.getWidth(), this.getHeight());
		super.paintComponent(g);
	}
	
	// Ensure that the preferred size is always square
	public Dimension getPreferredSize() {
		Dimension prefSize = super.getPreferredSize();
		prefSize.width = prefSize.height = Math.max(prefSize.width, prefSize.height);
		return prefSize;
	}
	
	// ### ACCESSORS ###
	
	public void setText(String text) {
		super.setText(text);
	}
}